// -*- coding: iso-8859-1-unix -*-

/**
* \file logger.cpp
*/

#include <log/logger.h>
#include <log/formatter/space_identation.h>
#include <log/formatter/print_level.h>
#include <log/formatter/add_breakline.h>

#include <boost/utility.hpp>

#include <functional>
#include <numeric>
#include <cstdio>

namespace logger {

logger::logger(std::auto_ptr<output_base> output)
  : level_(warning_level)
{
  outputs.push_back(output);

  add_formatter(formatter::make_space_identation());
  add_formatter(formatter::make_print_level());
  add_formatter(formatter::make_add_breakline());
}

logger::logger()
  : level_(warning_level)
{
  add_formatter(formatter::make_space_identation());
  add_formatter(formatter::make_print_level());
  add_formatter(formatter::make_add_breakline());
}

logger::logger(level l)
  : level_(l)
{
  add_formatter(formatter::make_space_identation());
  add_formatter(formatter::make_print_level());
  add_formatter(formatter::make_add_breakline());
}

scope_token logger::create_scope(const char* label)
{
  if(!scope_stack.get())
  {
    scope_stack.reset(new std::list<scope_info>);
    scope_stack->push_back(scope_info(""));
  }

  scope_stack->push_back(scope_info(label));
  scope_token token (boost::prior(scope_stack->end()));
  return token;
}

namespace {

struct scope_logical_and
{
  bool operator()(bool const& lhs, scope_info const& rhs) const
  {
    return lhs && rhs.deleted;
  }
  bool operator()(scope_info const& lhs, bool const& rhs) const
  {
    return lhs.deleted && rhs;
  }
};

}

void logger::remove_scope(scope_token const& token)
{
  token.scope_iterator->deleted = true;
  if(std::accumulate(token.scope_iterator, scope_stack->end()
                     , true, scope_logical_and()))
  {
    scope_stack->erase(token.scope_iterator, scope_stack->end());
  }
}

std::size_t logger::identation_level(scope_token const& token) const
{
  return std::distance(scope_stack->begin(), token.scope_iterator);
}

namespace {

void log_aux(logger const& l, level lev, scope_token const& token, std::string& string
             , boost::ptr_vector<formatter_base> const& formatters
             , boost::ptr_vector<output_base> const& outputs)
{
  for(boost::ptr_vector<formatter_base>::const_iterator 
        first = formatters.begin(), last = formatters.end()
        ;first != last; ++first)
  {
    first->format(l, lev, token, string);
  }

  for(boost::ptr_vector<output_base>::const_iterator
        first = outputs.begin(), last = outputs.end()
        ;first != last; ++first)
  {
    first->print(string);
  }
}

}

#ifndef LOGGER_DISABLE_VARARGS
void logger::vlog(level l, scope_token const& token, const char* format, va_list va)
{
#ifndef LOGGER_DISABLE_THREADS
  boost::unique_lock<boost::mutex> lock(mutex);
#endif
  assert(l > 0 && l <= debug_level);
  if(level_ >= l)
  {
    std::string string(120, '\0');

    int size = 0;
    {
      do
      {
        if(size != 0) string.resize(string.size()*2);
        size = vsnprintf(&string[0], string.size(), format, va);
      }
      while(size >= 0 && size >= string.size());
    }

    if(size >= 0)
    {
      string.resize(size);

      log_aux(*this, l, token, string, formatters, outputs);
    }
  }
}

void logger::log(level l, scope_token const& token, const char* format, ...)
{
  va_list argptr;
  va_start(argptr, format);
  vlog(l, token, format, argptr);
}
#endif

void logger::log(level l, scope_token const& token, std::string string)
{
#ifndef LOGGER_DISABLE_THREADS
  boost::unique_lock<boost::mutex> lock(mutex);
#endif
  if(level_ >= l)
  {
    log_aux(*this, l, token, string, formatters, outputs);
  }  
}


}
